/*
    WikiArchive - Category and Archive extensions for JSPWiki
    
    Copyright (C) 2007 Jens Gulden (jgulden@cs.tu-berlin.de)
    JSPWiki copyright (c) Janne Jalkanen (Janne.Jalkanen@iki.fi)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package de.tuberlin.cs.flp.jspwiki;

import com.ecyrd.jspwiki.ReferenceManager;
import com.ecyrd.jspwiki.WikiEngine;
import com.ecyrd.jspwiki.WikiPage;
import com.ecyrd.jspwiki.event.*;
import com.ecyrd.jspwiki.providers.WikiPageProvider;
import de.tuberlin.cs.flp.jspwiki.util.*;
import java.util.*;

/**
 * Handles Categories in the Wiki. A category is a page which contains other pages in the sense that
 * <ul>
 * <li>each page belonging to the category contains at least one link to the category-page</li>
 * <li>each page belonging to the category contains one of the member-phrases, as stored in the wiki page <code>CategoriesMemberPhrases</code> (or as configured via <code>jspwiki.categories.memberPhrases.page</code>)</li>
 * <li>the category-page includes the plugin <code>de.tuberlin.cs.flp.jspwiki.plugin.Category</code> (and usually nothing more, i.e. the category-page does not contain links to its members in wiki markup, but links are displayed by the plugin)</li>
 * <li>the category-page is linked from the central <code>Categories</code> page (or as configured via <code>jspwiki.categories.page</code>)</li>
 * </ul>
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7e42]
 * @author Jens Gulden
 * @see de.tuberlin.cs.flp.jspwiki.plugin.Category
 * @see de.tuberlin.cs.flp.jspwiki.tags.CategoryIterator
 */
public class CategoryManager implements WikiEventListener {

/**
 * Image file suffices to identify an image by its filename.
 * (Simple hard-coded solution. Should be configurable or taken from the Wiki's general configuration.)
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d53]
 */
    public static final String CATEGORY_INITIAL_SUFFIX = "Initial";

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d4d]
 */
    public static final String CATEGORY_INFO_SUFFIX = "Info";

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d47]
 */
    public static final String CATEGORY_EDITINFO_SUFFIX = "EditInfo";

/**
 * Property holding the name of the central Categories page on which all
 * CategoryXxx pages are registered to become known as categories.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d41]
 */
    public static final String PROPERTY_CATEGORY_PAGE = "jspwiki.categories.page";

/**
 * Default name of the central Categories page.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d3b]
 */
    public static final String DEFAULT_CATEGORY_PAGE = "Categories";

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d35]
 */
    public static final String PROPERTY_CATEGORY_MEMBER_PHRASES_PAGE = "jspwiki.categories.memberPhrases.page";

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d2f]
 */
    public static final String DEFAULT_CATEGORY_MEMBER_PHRASES_PAGE = "CategoriesMemberPhrases";

/**
 * 
 * 
 * @poseidon-object-id [Ie4883fm112767c1b91mm34bf]
 */
    public static final long NEW_MEMBER_TIMEOUT_SECONDS = 60 * 60 * 24;

/**
 * <p>Represents ...</p>
 * 
 * 
 * @poseidon-object-id [I724519c4m1127a063bbemm4016]
 */
    protected static java.util.Map instancesByEngine = new HashMap();

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d29]
 */
    protected Map membersToBeCreatedByCategory;

/**
 * 
 * 
 * @poseidon-object-id [Ie4883fm112767c1b91mm348b]
 */
    protected Map membersToBeCreatedTimestamp;

/**
 * <p>Represents ...</p>
 * 
 * 
 * @poseidon-object-id [I724519c4m1127a063bbemm3f22]
 */
    protected Map existingCategoryMembersCacheByCategory;
/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d20]
 */
    protected WikiEngine engine;

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d17]
 * @param engine 
 */
    protected  CategoryManager(WikiEngine engine) {        
        super();
        this.engine = engine;
        membersToBeCreatedByCategory = new HashMap();
        existingCategoryMembersCacheByCategory = new HashMap();
        membersToBeCreatedTimestamp = new HashMap();
        //engine.addWikiEventListener(this);
        PageEventFilter filter = new PageEventFilter();
        engine.getFilterManager().addPageFilter(filter, 0);
        WikiEventManager.addWikiEventListener(filter,this);        
    } 

/**
 * Static method to get an instance of CategoryManager.
 * 
 * This is the point where CategoryManager could be integrated into the
 * existing JSPWiki codebase by replacing this method with
 * engine.getCategoryManager().
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d10]
 * @return 
 * @param engine 
 */
    public static synchronized CategoryManager getInstance(WikiEngine engine) {        
        CategoryManager cm = (CategoryManager)instancesByEngine.get(engine);
        if (cm == null) {
            cm = new CategoryManager(engine);
            instancesByEngine.put(engine, cm);
        }
        return cm;
    } 

/**
 * Returns the name of the central Categories page on which all categories
 * are registered as links to the individual CategoryXxx pages.
 * (Note that the concept of categories is not applied recursively to declare
 * categories on the Categories page. Thus, a CategoryXxx page needs not
 * to link back to the main Categories page. Instead, there are links _from_
 * the Categories page _to_ the individual CategoryXxx pages.)
 * 
 * This can be configured in jspwiki.properties using jspwiki.categories.page. 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d0a]
 * @return 
 */
    public String getCategoryPageName() {        
        String categoryPageName = engine.getWikiProperties().getProperty(PROPERTY_CATEGORY_PAGE, DEFAULT_CATEGORY_PAGE);
        return categoryPageName;
    } 

/**
 * Returns the name of the central CategoriePhrasesPage, which contains
 * any number of possible textual phrases (seperated by blank lines in the
 * Wiki markup) that mark a page as being a member of a category.
 * 
 * This can be configured in jspwiki.properties using jspwiki.categories.memberPhrases.page. 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d04]
 * @return 
 */
    public String getCategoryMemberPhrasesPageName() {        
        String categoryMemberPhrasesPageName = engine.getWikiProperties().getProperty(PROPERTY_CATEGORY_MEMBER_PHRASES_PAGE, DEFAULT_CATEGORY_MEMBER_PHRASES_PAGE);
        return categoryMemberPhrasesPageName;
    } 

/**
 * Returns all member-phrases that mark a page as being a member of a
 * category.
 *  
 * @poseidon-object-id [Im7d594a32m112753776a4mm6cfe]
 * @return 
 */
    public Collection getCategoryMemberPhrases() {        
        Collection result = new ArrayList();
        String s = engine.getPureText( getCategoryMemberPhrasesPageName(), WikiPageProvider.LATEST_VERSION );
        if ( s != null ) {
            StringTokenizer st = new StringTokenizer(s, "\n", false);
            while ( st.hasMoreTokens() ) {
                String phrase = st.nextToken();
                phrase = phrase.trim();
                if ( phrase.length() > 0 ) {
                    result.add(phrase);                    
                }
            }
        }
        return result;
    } 

/**
 * Returns the default member-phrase that marks a page as being a member of a
 * category, also if no CategoriesMemberPhrases page (or accordingly configured
 * via jspwiki.category.memberPhrases.page).
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6cf8]
 * @return 
 */
    public String getDefaultCategoryMemberPhrase() {        
        Collection phrases = getCategoryMemberPhrases();
        if ( ! phrases.isEmpty() ) {
            return (String)getCategoryMemberPhrases().iterator().next();
        } else {
            return "";
        }
    } 

/**
 * Gets the members of a category which actually exist as saved pages.
 * 
 * @poseidon-object-id [I724519c4m1127a063bbemm3f62]
 * @param category 
 * @return 
 */
    protected java.util.Collection getExistingCategoryMembersUncached(String category) {        
        Collection referringPages = engine.getReferenceManager().findReferrers(category);
        if (referringPages == null) return Collections.EMPTY_LIST;
        Collection memberPhrases = getCategoryMemberPhrases();
        // filter all pages that link to the category page with a blank in link name
        ArrayList pages = new ArrayList();
        for (Iterator it = referringPages.iterator(); it.hasNext(); ) {
            String pname = (String)it.next();
            if ( pname.indexOf('/') == -1 ) { // WORKAROUND: somehow, ReferenceManager delivers some Attachments as referrer (??) (FileIcons/pdf.png)
                if ( ! ( pname.equals( category+CATEGORY_INITIAL_SUFFIX ) || pname.equals( category+CATEGORY_INFO_SUFFIX ) || pname.equals( category+CATEGORY_EDITINFO_SUFFIX ) ) ) {
                    String rawtext = engine.getPureText(pname, WikiPageProvider.LATEST_VERSION);
                    int found = -1;
                    for (Iterator jt = memberPhrases.iterator(); (found == -1) && jt.hasNext(); ) {
                        String phrase = (String)jt.next();
                        found = rawtext.indexOf(phrase);
                        if ( found != -1 ) {
                            pages.add(pname);
                        }
                    }
                }
            }
        }
        return pages;
    } 

/**
 * Returns all categories used in the wiki. Typically, this is a Collection
 * of Strings like CategoryXXX, CategoryYYY, CategoryZZZ etc.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6cf2]
 * @return 
 */
    public Collection getAllCategories() {        
        // get all links pointing from the Categories page to other pages
        ReferenceManager rm = engine.getReferenceManager();
        String pagename = getCategoryPageName();
        //WikiPage page = engine.getPage( pagename ); // make sure the page is indexed before we call rm.findRefersTo(..)
        String wikitext = engine.getPureText(pagename, WikiPageProvider.LATEST_VERSION);
        Collection links = WikiToolbox.parseWikiLinks(wikitext);
        return links;
    } 

/**
 * Gets the members of a category which actually exist as saved pages.
 * This is a cached operation, subsequent accesses to this are fast if
 * no page in the whole Wiki has been change in the meantime.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6ceb]
 * @param category 
 * @return 
 */
    public java.util.Collection getExistingCategoryMembers(String category) {        
        Collection existingCategoryMembersCache = (Collection)existingCategoryMembersCacheByCategory.get(category);
        if (existingCategoryMembersCache == null) {
            existingCategoryMembersCache = getExistingCategoryMembersUncached(category);
            existingCategoryMembersCacheByCategory.put(category, existingCategoryMembersCache);
        }
        return existingCategoryMembersCache;
    } 

/**
 * Gets all embers of a category, including those that have been registered as
 * yet-to-be-created members.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6ce4]
 * @param category 
 * @return 
 */
    public java.util.Collection getCategoryMembers(String category) {        
        Collection members = new ArrayList( getExistingCategoryMembers( category ) );
        members.addAll( getNewCategoryMembers( category ) );
        return members;
    } 

/**
 * Tests if a page is a category page.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6cdd]
 * @return 
 * @param page 
 */
    public boolean isCategory(String page) {        
        return ( ! getCategories(page).isEmpty() );
    } 

/**
 * Returns all categories in which the page is a member.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6cd6]
 * @return 
 * @param page 
 */
    public Collection getCategories(String page) {        
        Collection result = new ArrayList();
        for (Iterator it = getAllCategories().iterator(); it.hasNext(); ) {
            String category = (String)it.next();
            Collection members = getCategoryMembers(category);
            if (members.contains(page)) {
                result.add(category);
            }
        }
        return result;
    } 

/**
 * Adds a new member to the category. The page does not need to actually exist
 * in the Wiki yet.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6cce]
 * @param category 
 * @param newPage 
 */
    public void addCategoryMember(String category, String newPage) {        
    	Collection membersToBeCreated = membersToBeCreated(category); 
        membersToBeCreated.add(newPage);
        membersToBeCreatedTimestamp.put(newPage, new Date());
        WikiPage p = engine.getPage(newPage);
        engine.getSearchManager().reindexPage(p); // ???
    } 

/**
 * Gets those category members that have already been added to the category
 * but do not yet exist as saved pages. 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6cc7]
 * @param category 
 * @return 
 */
    public java.util.Collection getNewCategoryMembers(String category) {        
        // remove those that have been created in the meantime
        Collection pages = getExistingCategoryMembers(category);
        long now = (new Date()).getTime();
        Collection membersToBeCreated;
        synchronized (this) {
            membersToBeCreated = membersToBeCreated(category);
            Collection cc = new ArrayList( membersToBeCreated );
            for (Iterator it = cc.iterator(); it.hasNext(); ) {
            	String p = (String)it.next();
            	Date ts = (Date)membersToBeCreatedTimestamp.get(p);
            	// page has actually been created in the meantime, or it hasn't created at all and timeout is reached:
            	// in both cases remove from 'new-pages-to-be-created-as-members-of-this-category' list
            	if ( pages.contains(p) || ( now - ts.getTime() > NEW_MEMBER_TIMEOUT_SECONDS * 1000 )) {
                    membersToBeCreated.remove(p);
                    membersToBeCreatedTimestamp.remove(p);
            	}
            }
        }
        return membersToBeCreated;
    } 

/**
 * Gets those category members that have been added to the category
 * but are not yet tested whether they might in the meantime have been 
 * actually created as saved pages. 
 * 
 * @poseidon-object-id [I73403dc0m1127c422a9cmm3858]
 * @param category 
 * @return 
 */
    protected synchronized java.util.Collection membersToBeCreated(String category) {        
        Collection c = (Collection)membersToBeCreatedByCategory.get(category);
        if (c == null) {
            c = new ArrayList();
            membersToBeCreatedByCategory.put(category, c);
        }
        return c;
    } 

/**
 * Implementation of WikiEventListener.
 * 
 * This watches out for any change of a page and then invalidates the cached list of category members
 * (as it might have changed through the page change). Using the cache is a simple optimization.
 * 
 * @poseidon-object-id [I724519c4m1127a063bbemm3e23]
 * @param e 
 */
    public void actionPerformed(WikiEvent e) {        
    	if (e instanceof WikiPageEvent) {
    		switch (e.getType()) {
    			case WikiPageEvent.POST_SAVE:
    			case WikiPageEvent.PAGE_DELETED:
    				synchronized (existingCategoryMembersCacheByCategory) {
        				existingCategoryMembersCacheByCategory.clear(); // invalidate all category caches
					}
    		}
    	}
    } 

    /** @poseidon-generated */
    public WikiEngine getEngine() {
        return engine;
    }
    /** @poseidon-generated */
    public void setEngine(WikiEngine wikiEngine) {
        this.engine = wikiEngine;
    }
 }
