/*
    WikiArchive - Category and Archive extensions for JSPWiki
    
    Copyright (C) 2007 Jens Gulden (jgulden@cs.tu-berlin.de)
    JSPWiki copyright (c) Janne Jalkanen (Janne.Jalkanen@iki.fi)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package de.tuberlin.cs.flp.jspwiki.plugin;

import com.ecyrd.jspwiki.*;
import com.ecyrd.jspwiki.attachment.*;
import com.ecyrd.jspwiki.plugin.PluginException;
import com.ecyrd.jspwiki.plugin.WikiPlugin;
import com.ecyrd.jspwiki.providers.*;
import de.tuberlin.cs.flp.jspwiki.CategoryManager;
import de.tuberlin.cs.flp.jspwiki.util.*;
import de.tuberlin.cs.flp.jspwiki.util.WikiToolbox;
import java.io.*;
import java.util.*;

/**
 * <strong>[{Picture [src=page-or-attachment] [size=thumbnail|small|_medium_|large] [style=css-name] [textflow] }]</strong>
 * <p>Insert an image in the page using pre-set named size alternatives. The image can be taken from an image-archive's page. If a page is given as src, the first image inlined in the page will be used as picture, or the first image attachment to that page in case the page has no image inlined. If src is missing, the first image-attachment of the current page will be used. The names available for the size attribute (&quot;thumbnail&quot;, &quot;small&quot;, etc.) are configured in the jspwiki.properties using parameter jspwiki.picture.size.XXX=n.</p>
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7e33]
 * @author Jens Gulden
 */
public class Picture implements WikiPlugin {

/**
 * Prefix for all property entries used by this plugin.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm71d0]
 */
    protected static final String PROPERTIES_PREFIX = "jspwiki.picture";

/**
 * Executes the plugin.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm71c7]
 * @return HTML to be inserted into page
 * @param context The WikiContext.
 * @param params Parameters specified at plugin inclusion.
 * @throws PluginException if an error occurs
 */
    public String execute(WikiContext context, Map params) {        
        WikiEngine engine = context.getEngine();
        String src = (String)params.get("src");
        if (src == null) {
            src = context.getPage().getName();
        }
        String sizename = (String)params.get("size");
        if (sizename == null) {
            sizename = "default";
        }
        String style = (String)params.get("style");
        if (style == null) {
            style = "picture";
        }
        boolean textflow = params.containsKey("textflow");
        String align;
        if ( textflow ) {
            align = "left";
        } else {
            align = "";
        }
        
        return imageHTML(context, src, sizename, style, align, true, true);
    } 

/**
 * Returns HTML code displaying an image.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm71ba]
 * @return 
 * @param context 
 * @param src 
 * @param sizename 
 * @param style 
 * @param align 
 * @param clickToEnlarge 
 * @param description 
 */
    public static String imageHTML(WikiContext context, String src, String sizename, String style, String align, boolean clickToEnlarge, boolean description) {        
        WikiEngine engine = context.getEngine();
        try {
            // get maximum size in pixels
            Map props = engine.getWikiProperties();
            props = WikiToolbox.subMap(props, PROPERTIES_PREFIX+".size");
            int size;
            try {
                String sizeStr = (String)props.get(sizename);
                size = Integer.parseInt(sizeStr);
            } catch (NumberFormatException nfe) {
                return "unknown image size '"+sizename+"'";
            }
            // get image
            String html;
            Attachment att;
            AttachmentManager am = engine.getAttachmentManager();
            att = am.getAttachmentInfo(src); // full attachment name?
            if ( att == null ) { // direct attachment of current page?
                String newSrc = context.getPage().getName() + "/" + src;
                att = am.getAttachmentInfo(newSrc);
                if ( att != null ) {
                    src = newSrc;
                }
            }
            if ( att == null ) { // src is page name, not direct attachment (usual case)
                att = WikiToolbox.findImageAttachment(engine, src);
            }
            if ( att != null ) {
                String url = engine.getAttachmentURL(att.getName());
                InputStream in = am.getAttachmentStream(att);
                int[] dim = WikiToolbox.determineImageSize(in);
                int w = dim[0];
                int h = dim[1];
                if ( w > size ) {
                    float f = (float)size / w;
                    w *= f;
                    h *= f;
                }
                if ( h > size ) {
                    float f = (float)size / h;
                    w *= f;
                    h *= f;
                }
                String title;
                if ( clickToEnlarge ) {
                    title = "click to enlarge";
                } else {
                    title = src;
                }
                html = "<img class='"+style+"' src='"+url+"' width='"+w+"' height='"+h+"' align='"+align+"' title='"+title+"' alt='"+src+"' longdesc='"+engine.getViewURL(src)+"'/>";
                if ( clickToEnlarge && ( (w < dim[0]) || (h < dim[1]) ) ) {
                    html = "<a href='"+url+"' target='_blank' onclick='window.open(\""+url+"\", \""+src+"\", \"width="+(dim[0]+20)+",height="+(dim[1]+20)+",toolbar=no,location=no,scrollbars=yes,resizable=yes,status=no\");return false'>" +  html + "</a>";
                }
                if (description) {
                    String pageUrl = engine.getViewURL(src);
                    html += "<br/><small class='"+style+"'><i class='"+style+"'>Picture: </i><a href='"+pageUrl+"' class='"+style+"'>"+src+"</a></small><br/>";
                }
            
            } else {
                html = "<i>(image not found: '"+src+"')</i> ";
            }
            return html;
        } catch (Exception e) {
            return "<i>(no image: "+e.getMessage()+")</i> ";
        }
    } 
 }
