/*
    WikiArchive - Category and Archive extensions for JSPWiki
    
    Copyright (C) 2007 Jens Gulden (jgulden@cs.tu-berlin.de)
    JSPWiki copyright (c) Janne Jalkanen (Janne.Jalkanen@iki.fi)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package de.tuberlin.cs.flp.jspwiki.plugin.wikiarchive;

import com.ecyrd.jspwiki.*;
import de.tuberlin.cs.flp.jspwiki.CategoryManager;
import de.tuberlin.cs.flp.jspwiki.plugin.*;
import java.io.*;
import java.util.*;

/**
 * Generic base implementation for concrete ArchiveIO-classes. This makes implementing an <code>ArchiveIOHandler</code> easier.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7e3b]
 * @author Jens Gulden
 */
public abstract class ArchiveHandlerAbstract implements ArchiveHandler {
/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7567]
 */
    protected WikiEngine engine;

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7560]
 */
    protected Map config;

/**
 * Initialize this ArchiveHandler.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7558]
 * @param engine 
 * @param config 
 */
    public void init(WikiEngine engine, Map config) {        
        this.engine = engine;
        this.config = config;
    } 

/**
 * Perform either input from the <code>Archive</code> model interface
 * into the data source handled by this ArchiveIOHandler.
 * @poseidon-object-id [Im7d594a32m112753776a4mm754d]
 * @param plugin The WikiArchive plugin.
 * 
 * @return 
 * @param io Switch between input or output mode, <emp>true</emp>
 * for input into data source, <emp>false</emp>
 * for output into <code>Archive</code> model interface.
 * @param archive The <code>Archive</code> model interface.
 * @param data The data source handled by this ArchiveIOHandler.
 * @param mode The processing mode. This is a bit mask with:<br>
 * bit #0 set: update exisiting entries with new values, if new data is different<br>
 * bit #1 set: insert new entries<br>
 * bit #2 set: ask individually for each entry (generate HTML checkbox-form first)
 * @param exceptions Collection of entry names that should be treated as exceptions to the behaviour specified by <code>mode</code>. (Not used.)
 * @throws IOException if an i/o error occurs
 */
    public String performIO(boolean io, Archive archive, Object data, int mode, Collection exceptions) throws IOException {        
        boolean insert = (mode & WikiArchive.MODE_INSERT) != 0;
        boolean update = (mode & WikiArchive.MODE_UPDATE) != 0;
        boolean single = (mode & WikiArchive.MODE_SINGLE) != 0;
        StringBuffer html = new StringBuffer();
        Map all = parseAll(data);
        Map allIn = io ? all : archive;
        Map allOut = io ? archive : all;
        if (allIn.size()>0) {
            html.append("<table align='center' width='100%' bgcolor='#888888' cellpadding='1' cellspacing='0' border='0' bordercolor='#000000'><tr><td><table width='100%'>");
            html.append("<tr><td colspan='3' bgcolor='#eeeeee'><font face='Helvetica,Arial' size='-1'>").append(String.valueOf(allIn.size())).append(" Eintr&auml;ge:</font></td></tr>");
            boolean bright = false;
            for (Iterator it = allIn.entrySet().iterator(); it.hasNext(); ) {
                Map.Entry entry = (Map.Entry)it.next();
                String name = (String)entry.getKey();
                ArchiveEntry inEntry = (ArchiveEntry)entry.getValue();
                ArchiveEntry outEntry = (ArchiveEntry)allOut.get(name);
                html.append("\n<tr bgcolor='"+(bright?"#dddddd":"#aaaaaa")+"'>");
                if (outEntry==null) {
                    if ( (insert) && ( (!single)||((exceptions!=null)&&exceptions.contains(name)))) {
                        if (io) {
                            archive.add(inEntry);
                        } else {
                            createEntry(data, inEntry);
                            html.append(msgDone(inEntry, null, WikiArchive.MODE_INSERT));
                        }
                    } else { // not inserted, although new
                        if (single) {
                            html.append(msgNotDone(inEntry, null, WikiArchive.MODE_INSERT|WikiArchive.MODE_SINGLE));
                        } else {
                            html.append(msgNotDone(inEntry, null, WikiArchive.MODE_INSERT));
                        }
                    }
                    html.append(msgToDo(inEntry, null, WikiArchive.MODE_INSERT));
                } else { // outEntry exists
                    if ( ! equalEntries(inEntry, outEntry) ) {
                        if ( (update) && ( (!single) ||((exceptions!=null)&&exceptions.contains(name)))) {
                            if (io) {
                                archive.remove(name);
                                archive.add(inEntry);
                            } else {
                                replaceEntry(data, inEntry, outEntry);
                                html.append(msgDone(inEntry, outEntry, WikiArchive.MODE_UPDATE));
                            }
                        } else {
                            if (single) {
                                html.append(msgNotDone(inEntry, outEntry, WikiArchive.MODE_UPDATE|WikiArchive.MODE_SINGLE));
                            } else {
                                html.append(msgNotDone(inEntry, outEntry, WikiArchive.MODE_UPDATE));
                            }
                        }
                        html.append(msgToDo(inEntry, outEntry, WikiArchive.MODE_UPDATE));
                    } else {
                        html.append(msgToDo(inEntry, outEntry, WikiArchive.MODE_NONE));
                    }
                }
                html.append("</tr>");
            }
            html.append("</table></td></tr></table>");
        }
        return html.toString();
    } 

/**
 * Creates an empty data source to be filled for export.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7540]
 * @return 
 */
    public abstract Object createData() throws IOException;

/**
 * Converts the handled data source to string for export.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7538]
 * @return 
 * @param data 
 */
    public String dataToString(Object data) throws IOException {        
        return data.toString(); // default implementation, to be overwritten by subclasses
    } 

/**
 * Converts a string to a handled data source for import.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm752f]
 * @param s 
 * @param data 
 */
    public void stringToData(String s, Object data) throws IOException {        
        // ! assumes that data instanceof StringBuffer! MUST be overwritten by subclasses if different data-type is used
        StringBuffer sb = (StringBuffer)data;
        sb.append(s);
    } 

/**
 * Content type for export via HTTP.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7528]
 * @return 
 */
    public abstract String getContentSuffix();

/**
 * Content type for export via HTTP.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7522]
 * @return 
 */
    public abstract String getContentType();

/**
 * Retrieves all archive entries from the data source.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm751b]
 * @return 
 * @param data The data source handled by this ArchiveIOHandler.
 * @throws IOException if an i/o error occurs
 */
    protected abstract Map parseAll(Object data) throws IOException;

/**
 * Creates a new archive entry in the data source handled by this ArchiveIOHandler.
 * @poseidon-object-id [Im7d594a32m112753776a4mm7510]
 * @param plugin The WikiArcive plugin.
 * 
 * @param data The data source handled by this ArchiveIOHandler.
 * @param entry The entry.
 * @throws IOException if an i/o error occurs
 */
    protected abstract void createEntry(Object data, ArchiveEntry entry) throws IOException;

/**
 * Replaces an archive entry in the data source handled by this ArchiveIOHandler.
 * @poseidon-object-id [Im7d594a32m112753776a4mm7503]
 * @param plugin The WikiArcive plugin.
 * 
 * @param data The data source handled by this ArchiveIOHandler.
 * @param newEntry The new entry.
 * @param oldEntry The old entry.
 * @throws IOException if an i/o error occurs
 */
    protected abstract void replaceEntry(Object data, ArchiveEntry newEntry, ArchiveEntry oldEntry) throws IOException;

/**
 * Tests whether all data carried by two entries is the same.
 * Note that this is different from using Comparators on ArchiveEntry
 * objects, because Comparators only look at entry-names, while this method
 * compares all entry-data.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm74f6]
 * @return <emp>true</emp> if both entries have equals content, <emp>false</emp> otherwise.
 * @param a One archive entry.
 * @param b Another archive entry.
 */
    protected boolean equalEntries(ArchiveEntry a, ArchiveEntry b) {        
        return a.equals(b);
    } 

/**
 * Outputs a message informing about the possible action to be taken for an input entry, e.g.
 * whether it would be added as a new entry, or an existing one would be updated.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm74ea]
 * @return The message string. (Actually used to build a part of an HTML-form user interface.)
 * @param inEntry The input entry.
 * @param outEntry The currently existing possible output entry, or null if such does not exist yet.
 * @param mode The action mode, like the <code>mode</code> parameter of performIO().
 */
    protected abstract String msgToDo(ArchiveEntry inEntry, ArchiveEntry outEntry, int mode);

/**
 * Outputs a message informing about the action that has been taken for an entry.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm74dd]
 * @return The message string. (Actually used to build a part of an HTML-form user interface.)
 * @param entry The entry processed.
 * @param outEntry The currently existing possible output entry, or null if such does not exist yet.
 * @param mode The action mode, like the <code>mode</code> parameter of performIO().
 */
    protected abstract String msgDone(ArchiveEntry entry, ArchiveEntry outEntry, int mode);

/**
 * Outputs a message informing that an action has not been taken for an entry.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm74d0]
 * @return The message string. (Actually used to build a part of an HTML-form user interface.)
 * @param entry The entry not processed.
 * @param outEntry The currently existing possible output entry, or null if such does not exist yet.
 * @param mode The action mode, like the <code>mode</code> parameter of performIO().
 */
    protected abstract String msgNotDone(ArchiveEntry entry, ArchiveEntry outEntry, int mode);

    /** @poseidon-generated */
    public WikiEngine getEngine() {
        return engine;
    }
    /** @poseidon-generated */
    public void setEngine(WikiEngine wikiEngine) {
        this.engine = wikiEngine;
    }
 }
