/*
    WikiArchive - Category and Archive extensions for JSPWiki
    
    Copyright (C) 2007 Jens Gulden (jgulden@cs.tu-berlin.de)
    JSPWiki copyright (c) Janne Jalkanen (Janne.Jalkanen@iki.fi)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package de.tuberlin.cs.flp.jspwiki.plugin.wikiarchive;

import com.ecyrd.jspwiki.*;
import de.tuberlin.cs.flp.jspwiki.CategoryManager;
import de.tuberlin.cs.flp.jspwiki.util.*;
import java.io.*;
import java.util.*;

/**
 * I/O-handler for bidirectional data interchange from/to a BibTeX bibliography file.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7e3d]
 * @author Jens Gulden
 */
public class ArchiveHandlerBibTeX extends ArchiveHandlerExternalFormatAbstract {

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm78ba]
 */
    public static final String CONTENT_TYPE = "application/x-bibtex; charset=ISO-8859-1";

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm78b4]
 */
    public static final String CONTENT_SUFFIX = ".bib";

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm78ae]
 * @see #formats
 */
    protected static Properties DEFAULT_FORMATS;

/**
 * Default BibTeX-entries to labels mapping.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm78a5]
 * @see #entries
 */
    protected static Properties DEFAULT_ENTRIES;

/**
 * Default entries to options-map mapping.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm789f]
 * @see #entriesOptions
 */
    protected static Map DEFAULT_ENTRIES_OPTIONS;

/**
 * Default BibTeX-type to labels mapping.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7899]
 * @see #types
 */
    protected static Properties DEFAULT_TYPES;

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7893]
 */
    protected Properties types = null;

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm788d]
 */
    protected Properties entries = null;

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7887]
 */
    protected Map entriesOptions = null;

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7881]
 */
    protected Properties formats = null;

/**
 * Init.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7878]
 * @param engine 
 * @param config 
 */
    public void init(WikiEngine engine, Map config) {        
        super.init(engine, config);
        // read BibTeX types from Properties
        String typesFilename = (String)config.get("types");
        if ( typesFilename != null ) {
            try {
                FileInputStream in = new FileInputStream( typesFilename );
                types = new PropertiesOrdered();
                types.load( in );
                in.close();
            } catch (IOException ioe) {
                //nop, types remains null
            }
        }
        String entriesFilename = (String)config.get("entries");
        if ( entriesFilename != null ) {
            try {
                FileInputStream in = new FileInputStream( entriesFilename );
                entries = new PropertiesOrdered();
                entries.load( in );
                in.close();
                entriesOptions = new TreeMap();
                for (Iterator it = entries.entrySet().iterator(); it.hasNext(); ) {
                    Map.Entry e = (Map.Entry)it.next();
                    String val = (String)e.getValue();
                    int pos = val.indexOf(';'); // there are options following
                    if (pos!=-1) {
                        Properties options = new Properties();
                        String key = (String)e.getKey();
                        String opts = val.substring(pos+1);
                        val = val.substring(0, pos).trim();
                        StringTokenizer st = new StringTokenizer(opts, ";=", true);
                        while (st.hasMoreTokens()) {
                            String opt = st.nextToken();
                            String optval = null;
                            if (st.hasMoreTokens() && st.nextToken().equals("=") && st.hasMoreTokens()) {
                                optval = st.nextToken();
                            } else {
                                optval = "";
                            }
                            options.put(opt, optval);
                            if (st.hasMoreTokens()) {
                                String tok = st.nextToken();
                                if (!tok.equals(";")) {
                                    //throw new PluginException("illegal BibTeX-option for entry "+key+": "+tok+"' (expected ';')");
                                    //nop, non-fatal, TODO: log something
                                }
                            }
                        }
                        entriesOptions.put(key, options);
                        entries.put(key, val); // re-put with only val
                    }
                }
            } catch (IOException ioe) {
                //nop, types remains null
            }
        }
        String formatsFilename = (String)config.get("formats");
        if ( formatsFilename != null ) {
            try {
                FileInputStream in = new FileInputStream( formatsFilename );
                formats = new PropertiesOrdered();
                formats.load( in );
                in.close();
            } catch (IOException ioe) {
                //nop, types remains null
            }
        }
    } 

/**
 * Perform either input from the <code>Archive</code> model interface
 * into the data source handled by this ArchiveIOHandler.
 * @poseidon-object-id [Im7d594a32m112753776a4mm786a]
 * @param plugin The WikiArchive plugin.
 * 
 * @return 
 * @param io Switch between input or output mode, <emp>true</emp>
 * for input into data source, <emp>false</emp>
 * for output into <code>Archive</code> model interface.
 * @param archive The <code>Archive</code> model interface.
 * @param data The data source handled by this ArchiveIOHandler.
 * @param mode The processing mode. This is a bit mask with:<br>
 * bit #0 set: update exisiting entries with new values, if new data is different<br>
 * bit #1 set: insert new entries<br>
 * bit #2 set: ask individually for each entry (generate HTML checkbox-form first)
 * @param exceptions Collection of entry names that should be treated as exceptions to the behaviour specified by <code>mode</code>. (Not used.)
 * @throws IOException if an i/o error occurs
 */
    public String performIO(boolean io, Archive archive, Object data, int mode, Collection exceptions) throws IOException {        
        if (!io) {
            // when outputting, start with a header
            StringBuffer buf = (StringBuffer)data;
            String date = (new Date()).toString();
            buf.append("%% BibTeX generated by Bibliography-plugin for JSPWiki, " + date + ".\n\n");
        }
        String s = super.performIO(io, archive, data, mode, exceptions);
        return s;
    } 

/**
 * Content type for export via HTTP.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm785a]
 * @return 
 */
    public String getContentType() {        
        return CONTENT_TYPE;
    } 

/**
 * Content type for export via HTTP.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7854]
 * @return 
 */
    public String getContentSuffix() {        
        return CONTENT_SUFFIX;
    } 

/**
 * Creates a new archive entry in the data source handled by this ArchiveIOHandler.
 * @poseidon-object-id [Im7d594a32m112753776a4mm784c]
 * @param plugin The WikiArcive plugin.
 * 
 * @param data The data source handled by this ArchiveIOHandler.
 * @param entry The entry.
 */
    public void createEntry(Object data, ArchiveEntry entry) {        
        StringBuffer buf = (StringBuffer)data;
        ArchiveEntryBibTeX entryBibTeX = convertEntry(entry); //new ArchiveEntryBibTeX(this, (Map)entry); // ! dirty typecast from ArchiveEntry to Map
        buf.append(entry2bibtex(entryBibTeX));
    } 

/**
 * Replaces an archive entry in the data source handled by this ArchiveIOHandler.
 * @poseidon-object-id [Im7d594a32m112753776a4mm7841]
 * @param plugin The WikiArcive plugin.
 * 
 * @param data The data source handled by this ArchiveIOHandler.
 * @param newEntryA The new entry.
 * @param oldEntryA The old entry.
 */
    protected void replaceEntry(Object data, ArchiveEntry newEntryA, ArchiveEntry oldEntryA) {        
        ArchiveEntryBibTeX newEntry = (ArchiveEntryBibTeX)newEntryA;
        ArchiveEntryBibTeX oldEntry = (ArchiveEntryBibTeX)oldEntryA;
        StringBuffer buf = (StringBuffer)data;
        String old = entry2bibtex(oldEntry);
        int pos = buf.indexOf(old);
        if (pos != -1) {
            buf.delete(pos, pos+old.length());
            buf.insert(pos, entry2bibtex(newEntry));
        } else {
            createEntry(data, newEntry);
        }
    } 

/**
 * Retrieves all archive entries from the data source.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7837]
 * @return 
 * @param data The data source handled by this ArchiveIOHandler.
 * @throws IOException if an i/o error occurs
 */
    protected Map parseAll(Object data) throws IOException {        
        TreeMap result = new TreeMap();
        String bibtex = data.toString();
        bibtex = removeLineBreaksInQuotedStrings(bibtex); // workaround, StreamTokenizer cannot handle multi-line-strings
        bibtex = WikiToolbox.escape(bibtex, '\\'); // workaround: StreamTokenizer wants to interprete escape-sequences, so escape backslashes
        StreamTokenizer st = new StreamTokenizer(new StringReader(bibtex));
        syntax(st);
        st.nextToken();
        while ((st.ttype != '@') && (st.ttype != StreamTokenizer.TT_EOF)) { // allow anything between entries
            st.nextToken();
        }
        while (st.ttype == '@') { // != StreamTokenizer.TT_EOF) {
            st.nextToken();
            String type = st.sval;
            if (isType(type.toLowerCase())) {
                // parse single BibTeX entry
                st.nextToken();
                require(st, '{');
                st.nextToken();
                String name = until(st, ',');
                Map m = new PropertiesOrdered();
                while ( (st.ttype==',') && (st.nextToken()!='}') ) { // (allow ',' after last entry)
                    String key = until(st, '=');
                    st.nextToken();
                    String val;
                    if (st.ttype=='{') {
                        val = braceBlock(st);
                    } else {
                        switch (st.ttype) {
                            case '\"':
                            case StreamTokenizer.TT_WORD:
                                val = st.sval;
                                break;
                            case StreamTokenizer.TT_NUMBER: //(not used, parser recognizes numbers as strings)
                                double n = st.nval;
                                if (n == Math.round(n)) {
                                    val = String.valueOf((long)st.nval);
                                } else {
                                    val = String.valueOf(st.nval);
                                }
                                break;
                            default:
                                val = "### PARSE ERROR ###";
                        }
                    }
                    st.nextToken();
                    val = val.trim(); // need to trim because storage in wiki trims also, and we need exct string equality to determine whether whole entries are equals or not
                    m.put(key, val);
                    // (especially done for test data that has entries like 'month = "10~" # jan,')
                    while ((st.ttype!=StreamTokenizer.TT_EOF) && (st.ttype!=',') && (st.ttype!='}')) {
                        st.nextToken();
                    }
                }
                require(st, '}');
                m.put("type", WikiToolbox.capitalize(translateType( type.toLowerCase(), false )));
                String text = (String)m.remove("annote"); // special: wiki page text in field 'annote'
                m = translateKeys( m, false );
                /*for (Iterator it = getEntries().keySet().iterator(); it.hasNext(); ) { // make sure all entries (also empty ones) appear in the pages, unless they are marked optional
                }*/
                ArchiveEntryBibTeX entry = new ArchiveEntryBibTeX(this, WikiToolbox.normalizeName(name), m);
                entry.setText(text); // text may be null
                result.put(name, entry);
                
            } else { // ignore unknown blocks
                st.nextToken();
                braceBlock(st);
            }
            st.nextToken();
            while ((st.ttype != '@') && (st.ttype != StreamTokenizer.TT_EOF)) { // allow anything between entries
                st.nextToken();
            }
        }
        return result;
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm782d]
 * @return 
 * @param entry 
 */
    protected ArchiveEntryBibTeX convertEntry(ArchiveEntry entry) {        
        Map m = entry.getMap();
        Map newM = translateKeys( m, true );
        String type = (String)newM.get("type");
        newM.put("type", translateType( type.toLowerCase(), true ));
        ArchiveEntryBibTeX entryBibTeX = new ArchiveEntryBibTeX(this, entry.getName(), newM);
        entryBibTeX.setText(entry.getText()); // copy text (don't copy plaintext)
        return entryBibTeX;
    } 

/**
 * Get BibTeX-type to labels mapping. Keys of the properties-map are BibTeX type names (values for field 'type'),
 * values are (language-specific) label string that are used in Wiki pages.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7827]
 * @return Properties-map with BibTeX-type to labels mapping.
 * @see #types
 */
    public Properties getTypes() {        
        if ( types != null ) {
            return types;
        } else {
            return DEFAULT_TYPES;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm781f]
 * @return 
 */
    public Properties getFormats() {        
        if ( formats != null ) {
            return formats;
        } else {
            return DEFAULT_FORMATS;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7819]
 * @return 
 */
    public Properties getEntries() {        
        if ( entries != null ) {
            return entries;
        } else {
            return DEFAULT_ENTRIES;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7812]
 * @return 
 * @param entryName 
 */
    public Properties getEntryOptions(String entryName) {        
        Map m;
        if ( entriesOptions != null ) {
            m = entriesOptions;
        } else {
            m = DEFAULT_ENTRIES_OPTIONS;
        }
        if ( m != null ) {
            return (Properties)m.get( entryName );
        } else {
            return null;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm780b]
 * @return 
 * @param entryName 
 */
    public boolean isEntryOptional(String entryName) {        
        Properties options = getEntryOptions(entryName);
        if ( options != null ) {
            Object o = options.getProperty("optional");
            return ( o != null );
        } else {
            return false;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7804]
 * @return 
 * @param type The type.
 */
    public boolean isType(String type) {        
        return getTypes().containsKey(type);
    } 

/**
 * Converts the key names of Map m from/to label-strings.
 * This uses label strings specified as parameters.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm77fb]
 * @return Map with translated keys, same values.
 * @param m The map.
 * @param fromTo If <emp>true</emp>, convert from label to key, if <emp>false</emp>, convert from key to label.
 */
    protected Map translateKeys(Map m, boolean fromTo) {        
        if ( ! fromTo) {
            m = allEntries(m);
        }
        Map n = new PropertiesOrdered();
        for (Iterator it = m.entrySet().iterator(); it.hasNext(); ) {
            Map.Entry e = (Map.Entry)it.next();
            String key = (String)e.getKey();
            String newKey = translateKey(key, fromTo);
            String val = (String)e.getValue();
            String compKey = fromTo ? newKey : key;
            n.put(newKey, val);
        }
        if ( fromTo) {
            n = allEntries(n);
        }
        return n;
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm77f1]
 * @return 
 * @param m 
 */
    private Map allEntries(Map m) {        
        // make sure all entries (except optional ones) appear in map, even if empty
        Map n = new PropertiesOrdered();
        for (Iterator it = getEntries().keySet().iterator(); it.hasNext(); ) {
            String key = (String)it.next();
            String val = (String)m.get(key);
            if ( val == null ) {
                val = "";
            } else {
                val = val.trim();
            }
            if ( (val.length() > 0) || ( ! isEntryOptional(key) ) ) {
                n.put(key, val);
            }
        }
        return n;
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm77e9]
 * @return 
 * @param key 
 * @param fromTo 
 */
    protected String translateKey(String key, boolean fromTo) {        
        String newKey;
        if (fromTo) { // from
            newKey = WikiToolbox.keyByValue(getEntries(), key);
        } else {
            newKey = getEntries().getProperty(key);
        }
        if (newKey == null) {
            newKey = key; // no translation if not found
        }
        return newKey;
    } 

/**
 * Converts a BibTeX type name from or to its label-string.
 * This uses label string specified as parameters.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm77e1]
 * @return The translated string.
 * @param name The name or label.
 * @param fromTo If <emp>true</emp>, convert from label to name, if <emp>false</emp>, convert from name to label.
 * @see #types
 */
    protected String translateType(String name, boolean fromTo) {        
        if (name != null) {
            String n;
            if (!fromTo) { // TO Wiki, i.e. BibTeX->Wiki
                n = getTypes().getProperty(name);
            } else { // FROM Wiki, i.e. Wiki->BibTeX
                n = WikiToolbox.keyByValue(getTypes(), name);
            }
            if (n != null) {
                return n;
            }
        }
        return name; // no translation if not found in table
    } 

/**
 * Generate a single BibTeX entry string.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm77d6]
 * @return Single BibTeX entry.
 * @param entry The entry to output as BibTeX.
 */
    protected String entry2bibtex(ArchiveEntryBibTeX entry) {        
        StringBuffer sb = new StringBuffer();
        sb.append("@" + entry.getType() + "{" + entry.getName());
        for (Iterator it = getEntries().entrySet().iterator(); it.hasNext(); ) {
            Map.Entry e = (Map.Entry)it.next();
            String key = (String)e.getKey();
            if ( ! ( key.equals("type") || key.equals("annote") ) ) {
                String val = (String)entry.get(key);
                if ((val!=null) && (val.length()>0)) {
                    sb.append(bibtexEntry(key, val));
                }
            }
        }
        String text = entry.getText();
        if (text != null) {
            //text = text.replace('"', '\''); // simple: avoid any " being in the text
            sb.append(bibtexEntry("annote", text));
        }
        sb.append("\n}\n\n");
        return sb.toString();
    } 

/**
 * Generated a field entry string inside a BibTeX entry.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm77cc]
 * @return field entry string
 * @param key The field's name.
 * @param val The field's value.
 */
    protected static String bibtexEntry(String key, Object val) {        
        if (val!=null) {
            return ",\n"+key+"=\""+ WikiToolbox.escape((String)val, '\"') +"\"";
        } else {
            return "";
        }
    } 

/**
 * Tests if the tokenizer's current token matches a specified string.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm77be]
 * @param st The stream tokenizer.
 * @param tok The token to test match with.
 * @throws IOException If the token does not match.
 */
    protected static void require(StreamTokenizer st, String tok) throws IOException {        
        if (!st.sval.equals(tok)) {
            throw new IOException("expected '"+tok+"' at line "+st.lineno());
        }
    } 

/**
 * Tests if the tokenizer's current token matches a specified character.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm77b1]
 * @param st The stream tokenizer.
 * @param tok The token to test match with.
 * @throws IOException If the token does not match.
 */
    protected static void require(StreamTokenizer st, char tok) throws IOException {        
        if (st.ttype!=(int)tok) {
            throw new IOException("expected '"+tok+"' at line "+st.lineno());
        }
    } 

/**
 * Optionally allows a token to be the current token. If this token is the current one, skip to next token.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm77a5]
 * @param st The stream tokenizer.
 * @param c The token that may optionally occur.
 * @throws IOException if an i/o error occurs
 */
    protected static void optional(StreamTokenizer st, char c) throws IOException {        
        if (st.ttype==c) {
            st.nextToken();
        }
    } 

/**
 * Returns all tokens until another token occurs, concatenated as one string.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7799]
 * @return The concatenated tokens that appeared until the specified end token was reached.
 * @param st The stream tokenizer.
 * @param tok The token until which other tokens should be concatenated to the result string.
 * @throws IOException if an i/o error occurs
 */
    protected static String until(StreamTokenizer st, char tok) throws IOException {        
        StringBuffer sb = new StringBuffer();
        while ((st.ttype!=StreamTokenizer.TT_EOF)&&(st.ttype!=(int)tok)) {
            appendToken(st, sb);
            st.nextToken();
        }
        return sb.toString();
    } 

/**
 * Appends the current token to a StringBuffer.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7789]
 * @param st The stream tokenizer.
 * @param sb The string buffer.
 */
    protected static void appendToken(StreamTokenizer st, StringBuffer sb) {        
        switch (st.ttype) {
            case StreamTokenizer.TT_WORD:
                sb.append(st.sval);
                break;
            //case StreamTokenizer.TT_NUMBER:
            default:
                sb.append((char)st.ttype);
                break;
        }
    } 

/**
 * Parses a block enclosed in curly braces "{" and "}".
 * Returns the complete string inside the block. Nested curly braces are
 * possible and will be returned as part of the string.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7780]
 * @return The string enclosed in curly braces.
 * @param st The stream tokenizer.
 * @throws IOException if an i/o error occurs
 */
    protected static String braceBlock(StreamTokenizer st) throws IOException {        
        st.resetSyntax();
        StringBuffer sb = new StringBuffer();
        int cnt = 0;
        while ((cnt>=0)&&(st.nextToken()!=StreamTokenizer.TT_EOF)) {
            if (st.ttype==(int)'{') {
                cnt++;
            } else if (st.ttype==(int)'}') {
                cnt--;
            }
            if (cnt>=0) {
                sb.append((char)st.ttype);
            }
        }
        syntax(st);
        return sb.toString();
    } 

/**
 * Sets the tokenizer's syntax for parsing BibTeX.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7775]
 * @param st The stream tokenizer.
 */
    protected static void syntax(StreamTokenizer st) {        
        st.resetSyntax();
        st.whitespaceChars(0, 32);
        st.wordChars('0', '9');
        st.wordChars('A', 'Z');
        st.wordChars('a', 'z');
        st.quoteChar('\"');
        st.commentChar('%');
    } 

/**
 * Removes all line breaks inside quoted strings and replaces them with a single blank each.
 * This avoids quoted multi-line-strings, which are not not supported by StreamTokenizer.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm776d]
 * @return String without line-breaks inside quoted parts.
 * @param s The string.
 */
    private static String removeLineBreaksInQuotedStrings(String s) {        
        int len = s.length();
        char[] c = new char[len];
        s.getChars(0, len, c, 0);
        int startpos = findNextQuoteChar(c, 0);
        if (startpos != -1) {
            while (startpos != -1) {
                int endpos = findNextQuoteChar(c, startpos+1);
                if (endpos != -1) {
                    // overwrite linebreak chars in the range with blanks
                    for (int i = startpos+1; i<endpos; i++) {
                        if ((c[i]=='\n')||(c[i]=='\r')) {
                            c[i] = ' ';
                        }
                    }
                    startpos = findNextQuoteChar(c, endpos+1);
                } else {
                    startpos = -1;
                }
            }
            return new String(c);
        } else {
            return s; // unchanged
        }
    } 

/**
 * Finds the next occurrence of a quote char " inside a char[]-array,
 * starting from a specified position.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7762]
 * @return The next index-position of a quote occurrence, or -1 if no further quotes occur.
 * @param c The char array.
 * @param pos The start position.
 */
    private static int findNextQuoteChar(char[] c, int pos) {        
        while ((pos < c.length) && (c[pos] != '\"')) {
            pos++;
        }
        if (pos < c.length) {
            if (c[pos-1]=='\\') {
                pos = findNextQuoteChar(c, pos+1);
            }
            return pos;
        } else {
            return -1;
        }
    } 

/**
 * Puts an array of strings into the Properties-object,
 * buy using each array member both as key and value for each entry.
 * This is used for initializing dummy-properties.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7756]
 * @param p The properties.
 * @param s The string array.
 */
    protected static void putDefaults(Properties p, String[] s) {        
        for (int i=0; i<s.length; i++) {
            p.put(s[i], s[i]);
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm774f]
 */
    static {        
        DEFAULT_ENTRIES = new PropertiesOrdered();
        putDefaults( DEFAULT_ENTRIES, new String[] {
            "title",
            "author",
            "year",
            "type",
            "publisher",
            "address",
            "chapter",
            "pages", 
            "editor",
            "edition",
            "booktitle",
            "volume",
            "howpublished",
            "organization",
            "institution",
            "school",
            "month",
            "journal",
            "number",
            "series",
            "annote",
            "note",
            "custom1",
            "custom2",
            "custom3",
            "custom4",
            "custom5"
        });
        
        DEFAULT_ENTRIES_OPTIONS = new HashMap();
        Properties p = new Properties();
        p.put("optional", "");
        DEFAULT_ENTRIES_OPTIONS.put("custom1", p);
        DEFAULT_ENTRIES_OPTIONS.put("custom2", p);
        DEFAULT_ENTRIES_OPTIONS.put("custom3", p);
        DEFAULT_ENTRIES_OPTIONS.put("custom4", p);
        DEFAULT_ENTRIES_OPTIONS.put("custom5", p);
        
        DEFAULT_TYPES = new PropertiesOrdered();
        putDefaults( DEFAULT_TYPES, new String[] {
            "article",
            "book",
            "booklet",
            "conference",
            "inbook",
            "incollection",
            "inproceedings",
            "journal", 
            "manual",
            "masterthesis",
            "phdthesis",
            "proceedings",
            "techreport",
            "unpublished",
            "misc"
        });
        
        DEFAULT_FORMATS = new Properties();
        DEFAULT_FORMATS.put("default", "[%name] %author, %title, %publisher, %address, %year");
    }  
 }
