/*
    WikiArchive - Category and Archive extensions for JSPWiki
    
    Copyright (C) 2007 Jens Gulden (jgulden@cs.tu-berlin.de)
    JSPWiki copyright (c) Janne Jalkanen (Janne.Jalkanen@iki.fi)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package de.tuberlin.cs.flp.jspwiki.util;

import java.io.*;
import java.util.*;
import javax.servlet.http.*;

/**
 * Utility-class for handling multipart-form-data, used for uploading files with a web-browser.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7e44]
 * @author Jens Gulden
 */
public class MultipartFormData {

/**
 * Get POSTed file-upload data from request.
 * (It is a little shame that HttpRequest still does not support this on its own.)
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6f61]
 * @return Map containing key/value pairs sent via multipart/form-data, values are Strings.
 * Null if an error occurs or if the browser did not send a content-length entry.
 * @param request The request.
 */
    public static boolean isMultipartFormData(HttpServletRequest request) {        
        return request.getMethod().equals("POST") && request.getContentType().startsWith("multipart/form-data");
    } 

/**
 * Get POSTed file-upload data from request.
 * (HttpRequest still does not support this on its own.)
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6f58]
 * @return Map containing key/value pairs sent via multipart/form-data, values are Strings.
 * Null if an error occurs or if the browser did not send a content-length entry.
 * @param request The request.
 */
    public static Map getMultipartFormData(HttpServletRequest request) {        
        String encoding = request.getCharacterEncoding();
        return getMultipartFormData(request, encoding);
    } 

/**
 * Get POSTed file-upload data from request.
 * (HttpRequest still does not support this on its own.)
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6f4e]
 * @return Map containing key/value pairs sent via multipart/form-data, values are Strings.
 * Null if an error occurs or if the browser did not send a content-length entry.
 * @param request The request.
 * @param encoding the string encoding charset to use
 */
    public static Map getMultipartFormData(HttpServletRequest request, String encoding) {        
        HashMap result = new HashMap();
        int contentlength = request.getContentLength();
        if (contentlength != -1) {
            try {
                BufferedInputStream in = new BufferedInputStream(request.getInputStream());
                byte[] b=new byte[contentlength];
                int done = 0;
                while (done<contentlength) {
                    int read = in.read(b, done , contentlength-done);
                    done += read;
                    if (done<contentlength) {
                        try {
                            Thread.sleep(10);
                        } catch (InterruptedException ie) {
                            //nop
                        }
                    }
                }
                in.close();
                parseMultiPartFormData(b, result, encoding);
            } catch (IOException ioe) {
                return null;
            } catch (OutOfMemoryError oome) {
                return null;
            }
        } else {
            return null;
        }
        return result;
    } 

/**
 * Parse multipart-formdata from data sent via POST method.
 * Does not support multiple parameters with the same name (delivering an array).
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6f42]
 * @param b data bytes
 * @param parameters The parameters.
 * @param encoding The encoding.
 * @throws IOException if an i/o error occurs
 */
    public static void parseMultiPartFormData(byte[] b, Map parameters, String encoding) throws IOException {        
        int len = b.length;
        int pos = 0;
        // get separator
        while ((pos<len-2)&&(b[pos]!='\r')&&(b[pos+1]!='\n')) {
            pos++;
        }
        int sep = pos; // b[0..sep] is separator
        while ((pos<len-2)&&(!((b[pos]=='-')&&(b[pos+1]=='-')))) { // not at end (separator+"--")
            String paramname = null;
            pos+=2; // now on start of single-part header line
            while ((pos<len-2)&&(b[pos]!='\r'&&b[pos+1]!='\n')) { // another header line follows
                int linestart = pos;
                while ((pos<len-2)&&(b[pos]!='\r'&&b[pos+1]!='\n')) {
                    pos++;
                }
                // pos now on lineend
                String line;
                if (encoding != null) {
                    line = new String(b,linestart,pos-linestart, encoding);
                } else {
                    line = new String(b,linestart,pos-linestart);
                }
                if (line.startsWith("Content-Disposition: form-data;")) {
                    line=line.substring("Content-Disposition: form-data;".length());
                    Properties param = parseKeyValuePairs(line);
                    paramname=param.getProperty("name"); // trimmed
                    for (Enumeration e=param.keys();e.hasMoreElements();) { // add parameters like "filename" as <name>_filename=...
                        String key=(String)e.nextElement();
                        if (!key.equals("name")) {
                            String v = param.getProperty(key);
                            parameters.put(paramname+"_"+key,v);
                        }
                    }
                }
                // other single.part header lines are ignored (e.g. Content-Type)
                pos+=2; // to next line
            }
            // value now follows after \r\n:
            pos+=2;
            // fill StringBuffer until (incl.) separator
            int valstart=pos;
            int sepMatchCount=0;
            while ((pos<len-2)&&(sepMatchCount<sep)) {
                if (b[pos]==b[sepMatchCount]) {
                    sepMatchCount++;
                }
                else {
                    sepMatchCount=0;
                }
                pos++;
            }
            if (sepMatchCount==sep) {
                if (paramname!=null) {
                    String v;
                    if (encoding != null) {
                        v = new String(b,valstart,pos-valstart-sep-2, encoding); // -2 because '\r\n' before separator
                    } else {
                        v = new String(b,valstart,pos-valstart-sep-2); // -2 because '\r\n' before separator
                    }
                    putMulti(parameters, paramname, v);
                }
            }
        }
    } 

/**
 * Parse key/value pairs in a string. Key and value are sperated by '='.
 * the pairs are seperated by whitespace or ';'.
 * Multiple occurrences of the same key will result in a String[]-array
 * parse result for that key.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6f36]
 * @return Properties, with value entries either of type String or String[]
 * @param s The string to parse.
 */
    private static Properties parseKeyValuePairs(String s) {        
        // parses strings like: a="val"; b="val"; c="val"...
        StringTokenizer st = new StringTokenizer(s, " \t\n=\";", false);
        Properties p = new Properties();
        while (st.hasMoreTokens()) {
            String key = st.nextToken();
            if (st.hasMoreTokens()) {
                String val = st.nextToken();
                putMulti(p, key, val);
            }
        }
        return p;
    } 

/**
 * Puts a string value into a map, if an entry alreeady exists,
 * uses a string[]-array to collect multiple values.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6f2b]
 * @param m The map to put a key/value-pair into.
 * @param key The key.
 * @param val The value.
 */
    private static void putMulti(Map m, String key, String val) {        
        Object old = m.get(key);
        if (old != null) {
            String[] arr;
            if (old instanceof String) {
                arr = new String[2];
                arr[0] = (String)old;
                arr[1] = val;
            } else { // old instanceof String[]
                String[] oldS = (String[])old;
                arr = new String[oldS.length+1];
                System.arraycopy( oldS, 0, arr, 0, oldS.length );
                arr[oldS.length] = val;
            }
            m.remove(old);
            m.put(key, arr);
        } else {
            m.put(key, val);
        }
    } 
 }
