/*
    WikiArchive - Category and Archive extensions for JSPWiki
    
    Copyright (C) 2007 Jens Gulden (jgulden@cs.tu-berlin.de)
    JSPWiki copyright (c) Janne Jalkanen (Janne.Jalkanen@iki.fi)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package de.tuberlin.cs.flp.jspwiki.util;

import com.ecyrd.jspwiki.WikiEngine;
import com.ecyrd.jspwiki.WikiPage;
import com.ecyrd.jspwiki.attachment.Attachment;
import com.ecyrd.jspwiki.attachment.AttachmentManager;
import com.ecyrd.jspwiki.providers.ProviderException;
import com.ecyrd.jspwiki.providers.WikiPageProvider;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.*;
import javax.servlet.http.HttpServletRequest;
import org.devlib.schmidt.imageinfo.ImageInfo;

/**
 * Collection of tool functions used by package de.tuberlin.cs.flp.jspwiki.*.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm7e4d]
 * @author Jens Gulden
 */
public class WikiToolbox {

/**
 * Splits a string at the specified character occurrences.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6e51]
 * @return Collection of all substrings between the occurrences of <code>c</code>.
 * Contains a single element (whole <code>s</code>) if <code>c</code> is not contained in <code>s</code>.
 * @param s The string to be split.
 * @param c The character at the occurrences of which the splitting is performed.
 */
    public static Collection split(String s, char c) {        
        ArrayList l = new ArrayList();
        int pos = s.indexOf(c);
        int lastpos = 0;
        while (pos==lastpos && s.charAt(pos)==c) { // skip if starting with delimiter c
            pos = s.indexOf(c, pos+1);
            lastpos++;
        }
        while (pos!=-1) {
            l.add(s.substring(lastpos, pos));
            lastpos = pos+1;
            pos = s.indexOf(c, lastpos);
        }
        if (lastpos < s.length()) { // there is a rest
            l.add(s.substring(lastpos)); // add it
        }
        return l;
    } 

/**
 * Splits a string at the specified substring occurrences.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6e46]
 * @return Collection of all substrings between the occurrences of <code>c</code>.
 * Contains a single element (whole <code>s</code>) if <code>c</code> is not contained in <code>s</code>.
 * @param s The string to be split.
 * @param delim The substring at the occurrences of which the splitting is performed.
 */
    public static Collection split(String s, String delim) {        
        ArrayList l = new ArrayList();
        int len = delim.length();
        int pos = s.indexOf(delim);
        int lastpos = 0;
        while (pos == lastpos) { // skip if starting with delimiter
            lastpos += len;
            pos = s.indexOf(delim, pos + len);
        }
        while (pos != -1) {
            l.add(s.substring(lastpos, pos));
            lastpos = pos + len;
            pos = s.indexOf(delim, lastpos);
        }
        if ((lastpos+1) < s.length()) { // there is a rest
            l.add(s.substring(lastpos)); // add it
        }
        return l;
    } 

/**
 * Split at commas (',').
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6e3c]
 * @return 
 * @param s 
 */
    public static Collection split(String s) {        
        Collection c = new ArrayList();
        if ( s != null ) {
            StringTokenizer st = new StringTokenizer(s, ",", false);
            while ( st.hasMoreTokens() ) {
                String t = st.nextToken().trim();
                c.add(t);
            }
        }
        return c;
    } 

/**
 * Split at line-feed ('\n').
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6e35]
 * @return 
 * @param s 
 */
    public static Collection splitLines(String s) {        
        return split(s, '\n');
    } 

/**
 * Sets the first charcater in <code>s</code> to upper case.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6e2e]
 * @return Capitalized string.
 * @param s The string to be capitalized.
 */
    public static String capitalize(String s) {        
        if ((s!=null)&&(s.length()>0)) {
            return s.substring(0,1).toUpperCase()+s.substring(1);
        } else {
            return s;
        }
    } 

/**
 * Creates a string in which <code>c</code> is repeated <code>cnt</code> times.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6e24]
 * @return The string conatining the repeated character.
 * @param cnt Number of repetitions.
 * @param c The characater to be repeated.
 */
    public static String repeat(int cnt, char c) {        
        StringBuffer sb = new StringBuffer();
        for (int i=0; i<cnt; i++) {
            sb.append(c);
        }
        return sb.toString();
    } 

/**
 * Escapes all occurrences of <code>c</code> in <code>s</code> by a backslash.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6e19]
 * @return The escaped string.
 * @param s The string to be escaped.
 * @param c The character to be escaped.
 */
    public static String escape(String s, char c) {        
        int pos = s.indexOf(c);
        if (pos != -1) {
            int oldpos = 0;
            StringBuffer sb = new StringBuffer();
            while (pos != -1) {
                if (c!='\\' || s.charAt(pos+1)!='\"') { // special for backslashes: don't do this if it is an esacped quote char (backslash must not be last char in the string)
                    sb.append(s.substring(oldpos, pos));
                    sb.append('\\');
                    oldpos = pos;
                }
                pos = s.indexOf(c, pos+1);
            }
            sb.append(s.substring(oldpos));
            return sb.toString();
        } else {
            return s;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6e0d]
 * @return 
 * @param s 
 * @param search 
 * @param repl 
 */
    public static String replace(String s, String search, String repl) {        
        // simple
        int pos=s.indexOf(search);
        if (pos>=0) {
            return s.substring(0,pos)+repl+replace(s.substring(pos+search.length()),search,repl);
        } else {
            return s;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6e05]
 * @return 
 * @param map 
 * @param prefix 
 */
    public static Map subMap(Map map, String prefix) {        
        Map m = new HashMap();
        prefix = prefix + ".";
        int l = prefix.length();
        for ( Iterator it = map.entrySet().iterator(); it.hasNext(); ) {
            Map.Entry entry = (Map.Entry)it.next();
            String key = (String)entry.getKey();
            if ( key.startsWith(prefix) ) {
                key = key.substring(l);
                Object val = entry.getValue();
                m.put(key, val);
            }
        }
        return m;
    } 

/**
 * Gets the key of a property entry by the given value.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6dfd]
 * @return The key, or null if the value was not found.
 * @param p The properties.
 * @param v The value string.
 */
    public static String keyByValue(Properties p, String v) {        
        for (Iterator it = p.entrySet().iterator(); it.hasNext(); ) {
            Map.Entry e = (Map.Entry)it.next();
            String val = (String)e.getValue();
            if (v.equals(val)) {
                String key = (String)e.getKey();
                return key;
            }
        }
        return null;
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6df3]
 * @return 
 * @param in 
 */
    public static byte[] readInputStream(InputStream in) throws IOException {        
        ByteArrayOutputStream b = new ByteArrayOutputStream();
        int avail = in.available();
        while ( avail > 0 ) {
            byte[] buf = new byte[avail];
            in.read(buf);
            b.write(buf);
            avail = in.available();
        }
        return b.toByteArray();
    } 

/**
 * Creates a valid wiki page name from s.
 * I.e., blanks are removes and the string is converted to CamelCase.
 * ### Is there any corresponding function available in JSPWiki?
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6dea]
 * @return 
 * @param engine 
 * @param s 
 */
    public static String makePagename(WikiEngine engine, String s) {        
        s = toCamelCase(s);
        String enc = engine.getContentEncoding();
        try {
            s = URLEncoder.encode(s, enc);
            return s;
        } catch (UnsupportedEncodingException uee) {
            // nop
        }
        return s;
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6de3]
 * @return 
 * @param s 
 */
    public static List splitCamelCase(String s) {        
    	List l = new ArrayList();
    	int len  = s.length();
    	if (len == 0) return l;
    	int pos = 0;
    	do {
            int next = pos+1;
            boolean ok = false;
            while ( (next < len) && ( ! ok ) ) {
                char c = s.charAt(next);
                ok = Character.isUpperCase(c);
                if (!ok) {
                    next++;
                } 
            }
            String n = s.substring(pos, next);
            if (n.trim().length() > 0) {
                l.add(n);
            }
            pos = next;
    	} while (pos < len);
    	return l;
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6ddc]
 * @return 
 * @param s 
 */
    public static String toCamelCase(String s) {        
        StringTokenizer st = new StringTokenizer(s, " \t\r\n", false);
        StringBuffer sb = new StringBuffer();
        while ( st.hasMoreTokens() ) {
            String word = st.nextToken();
            word = capitalize(word);
            sb.append(word);
        }
        return sb.toString();
    } 

/**
 * Parses Wiki-markup and returns a collection of all links included in the page.
 * The links are capitalized (first letter upper case), in order to be
 * usable as page-names, which seem to be all uppercase in JSPWiki.
 * (I.e., all strings that are enclosed by "[]".)
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6dd5]
 * @return The names of all pages referenced by the Wiki-links in the page text.
 * @param pagedata The page text.
 */
    public static Collection parseWikiLinks(String pagedata) {        
        ArrayList l = new ArrayList();
        int pos = pagedata.indexOf('[');
        while (pos != -1) {
            int endpos = pagedata.indexOf(']', pos);
            if (endpos != -1) {
                if  (endpos > pos+1) {
                    String s = pagedata.substring(pos+1, endpos);
                    if (Character.isLetter(s.charAt(0))) { // make sure it's a real link, not plugin insertion etc.
                        s = Character.toUpperCase(s.charAt(0)) + s.substring(1);
                        l.add(s);
                    }
                }
                pos = pagedata.indexOf('[', endpos);
            } else {
                pos = -1;
            }
        }
        return l;
    } 

/**
 * Returns all Wiki-link targets included in a page.
 * These are all pages referenced by the specified page.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6dcb]
 * @return all links contained in the page
 * @param engine The WikiEngine
 * @param pagename name of the page to retrieve links from.
 */
    public static Collection parseWikiLinks(WikiEngine engine, String pagename) {        
        return parseWikiLinks(engine.getText(pagename));
    } 

/**
 * Parses a comma-seperated list of strings.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6dc1]
 * @return Collection with individual strings, each one trimmed.
 * @param s comma-seperated list of string
 */
    public static Collection parseList(String s) {        
        if (s == null) return null;
        StringTokenizer st = new StringTokenizer(s, ",", false);
        Collection result = new ArrayList();
        while (st.hasMoreTokens()) {
            String tok = st.nextToken();
            tok = tok.trim();
            result.add(tok);
        }
        return result;
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6db8]
 * @return 
 * @param s 
 */
    public static boolean parseBoolean(String s) {        
        return parseBoolean(s, false);
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6db0]
 * @return 
 * @param s 
 * @param deflt 
 */
    public static boolean parseBoolean(String s, boolean deflt) {        
        if ( s != null ) {
            return (s.equalsIgnoreCase("yes") || s.equalsIgnoreCase("true") || s.equalsIgnoreCase("on"));
        } else {
            return deflt;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6da8]
 * @return 
 * @param request 
 * @param name 
 */
    public static Collection parseMultiParameter(HttpServletRequest request, String name) {        
        
        String[] ss = request.getParameterValues(name);
        if ( ss != null ) {
            return Arrays.asList( ss );
        } else {
            return Collections.EMPTY_LIST;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d9f]
 * @return 
 * @param engine 
 * @param propertyName 
 * @param defaultPagename 
 */
    public static String getProperty(WikiEngine engine, String propertyName, String defaultPagename) {        
        return engine.getWikiProperties().getProperty(propertyName, defaultPagename);
    } 

/**
 * Makes sure that an entry's name can be used as a filename.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d98]
 * @return A normalized representation of the name.
 * @param name The entry name.
 */
    public static String normalizeName(String name) {        
        int len = name.length();
        char[] c = new char[len];
        name.getChars(0, len, c, 0);
        for (int i=0; i<len; i++) {
            if (!Character.isLetterOrDigit(c[i])) {
                c[i] = '_';
            }
        }
        String s = new String(c);
        s = capitalize(s);
        return s;
    } 

/**
 * Makes sure only '\n' is used as linebreak, not '\r\n'.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d8f]
 * @return 
 * @param s 
 */
    public static String normalizeLF(String s) {        
        
        s = replace(s, "\r\n", "\n");
        return s;
    } 

/**
 * Entries starting with % will be replaced by values in map.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d87]
 * @return String with %xxx-substrings resolved.
 * @param m The map.
 * @param f Template string including %xxx-substrings that will be replaced by map values.
 */
    public static String format(Map m, String f) {        
        int pos = f.indexOf('%');
        if (pos != -1) {
            int endpos = pos + 1;
            while ((endpos < f.length()) && Character.isLetterOrDigit(f.charAt(endpos))) {
                endpos++;
            }
            String key = f.substring(pos+1, endpos);
            String val = (String)m.get(key); // String or null
            return f.substring(0, pos) + ((val!=null)?val:"") +format(m, f.substring(endpos)); // recursion
        } else {
            return f;
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d7c]
 * @return 
 * @param s 
 * @param entryDelim 
 */
    public static Properties parseProperties(String s, String entryDelim) {        
        Properties result = new Properties();
        Collection entries = split(s, entryDelim);
        for (Iterator it = entries.iterator(); it.hasNext(); ) {
            String l = (String)it.next();
            int pos = l.indexOf('=');
            String key;
            String val;
            if (pos != -1) {
                key = l.substring(0, pos).trim();
                val = l.substring(pos + 1).trim();
            } else {
                key = l.trim();
                val = "";
            }
            result.put(key, val);
        }
        return result;
    } 

/**
 * Returns first image-attachment, or null if no attachment is an image or no attachments at all exist.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d74]
 * @return 
 * @param engine 
 * @param pagename 
 */
    public static Attachment findImageAttachment(WikiEngine engine, String pagename) {        
        WikiPage page = engine.getPage(pagename);
        if (page == null) return null;
        AttachmentManager am = engine.getAttachmentManager();
        try {
            Collection attachments = am.listAttachments(page);
            return findImageAttachment(engine, pagename, attachments);
        } catch (ProviderException pe) {
            return null;
        }
    } 

/**
 * Returns first image-attachment, or null if no attachment is an image or no attachments at all exist.
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d6b]
 * @return 
 * @param engine 
 * @param pagename 
 * @param attachments 
 */
    public static Attachment findImageAttachment(WikiEngine engine, String pagename, Collection attachments) {        
        Attachment last = null;
        for (Iterator at = attachments.iterator(); at.hasNext(); ) {
            Attachment a = (Attachment)at.next();
            String filename = a.getFileName();
            if (filename != null) {
                if ( isImage(engine, filename) ) {
                    last = a;
                    if ( engine.getPureText(pagename, WikiPageProvider.LATEST_VERSION).indexOf( "[" + filename + "]" ) != -1 ) { // make sure it's linked in the page
                        return a;
                    }
                }
            }
        }
        // reached here: maybe there was an image-attachment, but no link at all in the page        
        return last; // may be null
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d63]
 * @return 
 * @param in 
 */
    public static int[] determineImageSize(InputStream in) {        
        ImageInfo info = new ImageInfo();
        info.setInput(in);
        if ( info.check() ) {
            return new int[] { info.getWidth(), info.getHeight() };
        } else {
            return new int[] { 32, 32 }; // cannot determine
        }
    } 

/**
 * 
 * 
 * @poseidon-object-id [Im7d594a32m112753776a4mm6d5b]
 * @return 
 * @param engine 
 * @param name 
 */
    public static boolean isImage(WikiEngine engine, String name) {        
        Collection img = engine.getAllInlinedImagePatterns();
        for (Iterator it = img.iterator(); it.hasNext(); ) {
            String suffix = (String)it.next();
            suffix = suffix.substring(1); // assume all patterns start with '*', e.g. '*.png'
            if (name.endsWith(suffix)) {
                return true;
            }
        }
        return false;
    } 
 }
